//  *********************************
// Procedural texture map for a solid "F" shape
// **********************************

// ***************************
// Vertex Shader, named "vertexShader_P_MV_tex"
//     Sets the position and color of a vertex.
//   The projection and modelview matrices are used to position the vertex.
//   It copies the color to "theColor" so that the fragment shader can access it.
//   It copies the texture coordinates to "theTexCoords" so that the fragment shader can access them.
// ***************************
#beginglsl vertexshader vertexShader_P_MV_tex
#version 330 core
layout (location = 0) in vec3 vPos;	       // Position in attribute location 0
layout (location = 1) in vec3 vColor;      // Color in attribute location 1    
layout (location = 3) in vec2 vTexCoords;  // Color in attribute location 3    
out vec3 theColor;                         // output a color to the fragment shader    
out vec2 theTexCoords;                     // output a color to the fragment shader    
uniform mat4 projectionMatrix;             // The projection matrix    
uniform mat4 modelviewMatrix;              // The model-view matrix    
void main()    
{    
   gl_Position = projectionMatrix * modelviewMatrix * vec4(vPos.x, vPos.y, vPos.z, 1.0);    
   theColor = vColor; 
   theTexCoords = vTexCoords;
}
#endglsl 

// ***************************
// Fragment shader, named "fragmentShader_Fshape"
//    Set a general color using a fragment shader. (A "fragment" is a "pixel".)
//    The color value is passed in, obtained from the colors on the vertices.
//    Color values range from 0.0 to 1.0.
//    First three values are Red/Green/Blue (RGB).
//    Fourth color value (alpha) is 1.0, meaning there is no transparency.
// ***************************
#beginglsl fragmentshader fragmentShader_Fshape
#version 330 core    
in vec3 theColor;        // Color value came from the vertex shader (smoothed)   
in vec2 theTexCoords;	 // Texture coordinates (smoothed)
out vec4 FragColor;      // Color that will be used for the fragment    

bool InFshape( vec2 pos );	// Function to recognize interior of "F" shape

void main()    
{   
	vec2 wrappedTexCoords = fract(theTexCoords);	// Wrap s,t to [0,1].
	if ( InFshape(wrappedTexCoords) ) {		
		FragColor = vec4( 0, 0, 0, 1 );         // Black color inside the "F"
	}
	else {
        FragColor = vec4(theColor, 1.0f);   // Add alpha value of 1.0.
	}
}
#endglsl

// *******************************
// Recognize the interior of an "F" shape
//   Input "pos" contains s,t  texture coordinates.
//   Returns: true if inside the "F" shape.
//            false otherwise
// ******************************
#beginglsl codeblock inFshape
bool InFshape( vec2 pos ) {
	float sideMargin = 0.2;		// Left-to-right, F is in [sideMargin, 1-sideMargin]
	float verticalMargin = 0.1;	// Bottom-to-top, F is in [vertMargin, 1-vertMargin]
	float postWidth = 0.2;      // Width of the F's post
	float armWidth = 0.2;       // Width of the F's arms
	if ( pos.x<sideMargin || pos.x>1.0-sideMargin ||
	         pos.y<verticalMargin || pos.y>1.0-verticalMargin ) {
		 return false;
    }
	if ( pos.x<=sideMargin+postWidth || pos.y >= 1.0-verticalMargin-armWidth ) {
	    return true;
	}
	return ( pos.y <= 0.5+0.5*armWidth && pos.y >= 0.5-0.5*armWidth );
}
#endglsl

// **************************
// Distance from NxN grid points
// **************************
#beginglsl codeblock distNNgrid
float DistNbyNgrid( vec2 pos, int N ) {
	pos = N*pos;
	vec2 distUp = fract(pos);
	vec2 distDown = fract(-pos);
	vec2 distMins = min(distUp, distDown);
	return length(distMins)/N;
}
#endglsl

// ***************************
// Fragment shader, named "fragmentShader_FshapeNN"
//    Set a general color using a fragment shader. (A "fragment" is a "pixel".)
//    The color value is passed in, obtained from the colors on the vertices.
//    Color values range from 0.0 to 1.0.
//    First three values are Red/Green/Blue (RGB).
//    Fourth color value (alpha) is 1.0, meaning there is no transparency.
// ***************************
#beginglsl fragmentshader fragmentShader_FshapeNN
#version 330 core    
in vec3 theColor;        // Color value came from the vertex shader (smoothed)   
in vec2 theTexCoords;	 // Texture coordinates (smoothed)
out vec4 FragColor;      // Color that will be used for the fragment    

bool InFshape( vec2 pos );	// Function to recognize interior of "F" shape
float DistNbyNgrid( vec2 pos, int N );

void main()    
{   
	vec2 wrappedTexCoords = fract(theTexCoords);	// Wrap s,t to [0,1].
	if ( InFshape(wrappedTexCoords) ) {		
		FragColor = vec4( 0, 0, 0, 1 );         // Black color inside the "F"
	}
	else {
        FragColor = vec4(theColor, 1.0f);   // Add alpha value of 1.0.
	}
	float dist = DistNbyNgrid(theTexCoords, 6) * 6 / 0.25;
	if ( dist<1 ) {
	    FragColor = mix( vec4( 0, 1, 0.4, 1), FragColor, dist );  // Blend with green based on distance
	}
}
#endglsl
